@@ -1,4 +1,4 @@
-""" Handy utility functions."""
+""" Handy utility functions. """
 
 from __future__ import annotations
 
@@ -10,11 +10,16 @@ import json.decoder
 import os
 import pkgutil
 import random
+import re
 import sys
+import threading
+import time
+import typing
 import warnings
 from contextlib import contextmanager
 from distutils.version import StrictVersion
 from enum import Enum
+from io import BytesIO
 from numbers import Number
 from pathlib import Path
 from typing import (
@@ -27,16 +32,20 @@ from typing import (
     NewType,
     Tuple,
     Type,
+    TypeVar,
+    Union,
 )
 
 import aiohttp
 import fsspec.asyn
 import httpx
+import matplotlib.pyplot as plt
 import requests
 from pydantic import BaseModel, Json, parse_obj_as
 
 import gradio
 from gradio.context import Context
+from gradio.strings import en
 
 if TYPE_CHECKING:  # Only import for type checking (is False at runtime).
     from gradio.blocks import BlockContext
@@ -46,13 +55,18 @@ analytics_url = "https://api.gradio.app/"
 PKG_VERSION_URL = "https://api.gradio.app/pkg-version"
 JSON_PATH = os.path.join(os.path.dirname(gradio.__file__), "launches.json")
 
+T = TypeVar("T")
+
 
 def version_check():
     try:
-        current_pkg_version = (
-            pkgutil.get_data(__name__, "version.txt").decode("ascii").strip()
-        )
-        latest_pkg_version = requests.get(url=PKG_VERSION_URL).json()["version"]
+        version_data = pkgutil.get_data(__name__, "version.txt")
+        if not version_data:
+            raise FileNotFoundError
+        current_pkg_version = version_data.decode("ascii").strip()
+        latest_pkg_version = requests.get(url=PKG_VERSION_URL, timeout=3).json()[
+            "version"
+        ]
         if StrictVersion(latest_pkg_version) > StrictVersion(current_pkg_version):
             print(
                 "IMPORTANT: You are using gradio version {}, "
@@ -71,54 +85,83 @@ def version_check():
 
 
 def get_local_ip_address() -> str:
-    try:
-        ip_address = requests.get("https://api.ipify.org", timeout=3).text
-    except (requests.ConnectionError, requests.exceptions.ReadTimeout):
-        ip_address = "No internet connection"
+    """Gets the public IP address or returns the string "No internet connection" if unable to obtain it. Does not make a new request if the IP address has already been obtained."""
+    if Context.ip_address is None:
+        try:
+            ip_address = requests.get(
+                "https://checkip.amazonaws.com/", timeout=3
+            ).text.strip()
+        except (requests.ConnectionError, requests.exceptions.ReadTimeout):
+            ip_address = "No internet connection"
+        Context.ip_address = ip_address
+    else:
+        ip_address = Context.ip_address
     return ip_address
 
 
-def initiated_analytics(data: Dict[str:Any]) -> None:
-    try:
-        requests.post(
-            analytics_url + "gradio-initiated-analytics/", data=data, timeout=3
-        )
-    except (requests.ConnectionError, requests.exceptions.ReadTimeout):
-        pass  # do not push analytics if no network
+def initiated_analytics(data: Dict[str, Any]) -> None:
+    data.update({"ip_address": get_local_ip_address()})
+
+    def initiated_analytics_thread(data: Dict[str, Any]) -> None:
+        try:
+            requests.post(
+                analytics_url + "gradio-initiated-analytics/", data=data, timeout=3
+            )
+        except (requests.ConnectionError, requests.exceptions.ReadTimeout):
+            pass  # do not push analytics if no network
+
+    threading.Thread(target=initiated_analytics_thread, args=(data,)).start()
 
 
 def launch_analytics(data: Dict[str, Any]) -> None:
-    try:
-        requests.post(
-            analytics_url + "gradio-launched-analytics/", data=data, timeout=3
-        )
-    except (requests.ConnectionError, requests.exceptions.ReadTimeout):
-        pass  # do not push analytics if no network
+    data.update({"ip_address": get_local_ip_address()})
+
+    def launch_analytics_thread(data: Dict[str, Any]) -> None:
+        try:
+            requests.post(
+                analytics_url + "gradio-launched-analytics/", data=data, timeout=3
+            )
+        except (requests.ConnectionError, requests.exceptions.ReadTimeout):
+            pass  # do not push analytics if no network
+
+    threading.Thread(target=launch_analytics_thread, args=(data,)).start()
 
 
 def integration_analytics(data: Dict[str, Any]) -> None:
-    try:
-        requests.post(
-            analytics_url + "gradio-integration-analytics/", data=data, timeout=3
-        )
-    except (requests.ConnectionError, requests.exceptions.ReadTimeout):
-        pass  # do not push analytics if no network
+    data.update({"ip_address": get_local_ip_address()})
+
+    def integration_analytics_thread(data: Dict[str, Any]) -> None:
+        try:
+            requests.post(
+                analytics_url + "gradio-integration-analytics/", data=data, timeout=3
+            )
+        except (requests.ConnectionError, requests.exceptions.ReadTimeout):
+            pass  # do not push analytics if no network
+
+    threading.Thread(target=integration_analytics_thread, args=(data,)).start()
 
 
-def error_analytics(ip_address: str, message: str) -> None:
+def error_analytics(message: str) -> None:
     """
     Send error analytics if there is network
-    :param type: RuntimeError or NameError
+    :param ip_address: IP address where error occurred
+    :param message: Details about error
     """
-    data = {"ip_address": ip_address, "error": message}
-    try:
-        requests.post(analytics_url + "gradio-error-analytics/", data=data, timeout=3)
-    except (requests.ConnectionError, requests.exceptions.ReadTimeout):
-        pass  # do not push analytics if no network
+    data = {"ip_address": get_local_ip_address(), "error": message}
+
+    def error_analytics_thread(data: Dict[str, Any]) -> None:
+        try:
+            requests.post(
+                analytics_url + "gradio-error-analytics/", data=data, timeout=3
+            )
+        except (requests.ConnectionError, requests.exceptions.ReadTimeout):
+            pass  # do not push analytics if no network
+
+    threading.Thread(target=error_analytics_thread, args=(data,)).start()
 
 
-async def log_feature_analytics(ip_address: str, feature: str) -> None:
-    data = {"ip_address": ip_address, "feature": feature}
+async def log_feature_analytics(feature: str) -> None:
+    data = {"ip_address": get_local_ip_address(), "feature": feature}
     async with aiohttp.ClientSession() as session:
         try:
             async with session.post(
@@ -146,6 +189,12 @@ def colab_check() -> bool:
     return is_colab
 
 
+def kaggle_check() -> bool:
+    return bool(
+        os.environ.get("KAGGLE_KERNEL_RUN_TYPE") or os.environ.get("GFOOTBALL_DATA_DIR")
+    )
+
+
 def ipython_check() -> bool:
     """
     Check if interface is launching from iPython (not colab)
@@ -174,7 +223,7 @@ def readme_to_html(article: str) -> str:
 
 def show_tip(interface: gradio.Blocks) -> None:
     if interface.show_tips and random.random() < 1.5:
-        tip: str = random.choice(gradio.strings.en["TIPS"])
+        tip: str = random.choice(en["TIPS"])
         print(f"Tip: {tip}")
 
 
@@ -189,7 +238,7 @@ def launch_counter() -> None:
                 launches = json.load(j)
             launches["launches"] += 1
             if launches["launches"] in [25, 50, 150, 500, 1000]:
-                print(gradio.strings.en["BETA_INVITE"])
+                print(en["BETA_INVITE"])
             with open(JSON_PATH, "w") as j:
                 j.write(json.dumps(launches))
     except:
@@ -258,11 +307,12 @@ def assert_configs_are_equivalent_besides_ids(
     return True
 
 
-def format_ner_list(input_string: str, ner_groups: Dict[str : str | int]):
+def format_ner_list(input_string: str, ner_groups: List[Dict[str, str | int]]):
     if len(ner_groups) == 0:
         return [(input_string, None)]
 
     output = []
+    end = 0
     prev_end = 0
 
     for group in ner_groups:
@@ -275,7 +325,7 @@ def format_ner_list(input_string: str, ner_groups: Dict[str : str | int]):
     return output
 
 
-def delete_none(_dict, skip_value=False):
+def delete_none(_dict: T, skip_value: bool = False) -> T:
     """
     Delete None values recursively from all of the dictionaries, tuples, lists, sets.
     Credit: https://stackoverflow.com/a/66127889/5209347
@@ -302,7 +352,7 @@ def resolve_singleton(_list: List[Any] | Any) -> Any:
         return _list
 
 
-def component_or_layout_class(cls_name: str) -> Component | BlockContext:
+def component_or_layout_class(cls_name: str) -> Type[Component] | Type[BlockContext]:
     """
     Returns the component, template, or layout class with the given class name, or
     raises a ValueError if not found.
@@ -312,6 +362,7 @@ def component_or_layout_class(cls_name: str) -> Component | BlockContext:
     Returns:
     cls: the component class
     """
+    import gradio.blocks
     import gradio.components
     import gradio.layouts
     import gradio.templates
@@ -340,7 +391,7 @@ def component_or_layout_class(cls_name: str) -> Component | BlockContext:
     raise ValueError(f"No such component or layout: {cls_name}")
 
 
-def synchronize_async(func: Callable, *args, **kwargs):
+def synchronize_async(func: Callable, *args, **kwargs) -> Any:
     """
     Runs async functions in sync scopes.
 
@@ -394,19 +445,19 @@ def async_iteration(iterator):
         raise StopAsyncIteration()
 
 
-class Request:
+class AsyncRequest:
     """
-    The Request class is a low-level API that allow you to create asynchronous HTTP requests without a context manager.
-    Compared to making calls by using httpx directly, Request offers more flexibility and control over:
+    The AsyncRequest class is a low-level API that allow you to create asynchronous HTTP requests without a context manager.
+    Compared to making calls by using httpx directly, AsyncRequest offers more flexibility and control over:
         (1) Includes response validation functionality both using validation models and functions.
         (2) Since we're still using httpx.Request class by wrapping it, we have all it's functionalities.
         (3) Exceptions are handled silently during the request call, which gives us the ability to inspect each one
         individually in the case of multiple asynchronous request calls and some of them failing.
-        (4) Provides HTTP request types with Request.Method Enum class for ease of usage
-    Request also offers some util functions such as has_exception, is_valid and status to inspect get detailed
+        (4) Provides HTTP request types with AsyncRequest.Method Enum class for ease of usage
+    AsyncRequest also offers some util functions such as has_exception, is_valid and status to inspect get detailed
     information about executed request call.
 
-    The basic usage of Request is as follows: create a Request object with inputs(method, url etc.). Then use it
+    The basic usage of AsyncRequest is as follows: create a AsyncRequest object with inputs(method, url etc.). Then use it
     with the "await" statement, and then you can use util functions to do some post request checks depending on your use-case.
     Finally, call the get_validated_data function to get the response data.
 
@@ -437,10 +488,11 @@ class Request:
         method: Method,
         url: str,
         *,
-        validation_model: Type[BaseModel] = None,
-        validation_function: Callable = None,
+        validation_model: Type[BaseModel] | None = None,
+        validation_function: Union[Callable, None] = None,
         exception_type: Type[Exception] = Exception,
         raise_for_status: bool = False,
+        client: httpx.AsyncClient | None = None,
         **kwargs,
     ):
         """
@@ -454,8 +506,7 @@ class Request:
             exception_class(Type[Exception]): a exception type to throw with its type
             raise_for_status(bool): a flag that determines to raise httpx.Request.raise_for_status() exceptions.
         """
-        self._response = None
-        self._exception = None
+        self._exception: Union[Exception, None] = None
         self._status = None
         self._raise_for_status = raise_for_status
         self._validation_model = validation_model
@@ -464,14 +515,15 @@ class Request:
         self._validated_data = None
         # Create request
         self._request = self._create_request(method, url, **kwargs)
+        self.client_ = client or self.client
 
-    def __await__(self) -> Generator[None, Any, "Request"]:
+    def __await__(self) -> Generator[None, Any, "AsyncRequest"]:
         """
         Wrap Request's __await__ magic function to create request calls which are executed in one line.
         """
         return self.__run().__await__()
 
-    async def __run(self) -> Request:
+    async def __run(self) -> AsyncRequest:
         """
         Manage the request call lifecycle.
         Execute the request by sending it through the client, then check its status.
@@ -485,7 +537,7 @@ class Request:
         """
         try:
             # Send the request and get the response.
-            self._response: httpx.Response = await Request.client.send(self._request)
+            self._response: httpx.Response = await self.client_.send(self._request)
             # Raise for _status
             self._status = self._response.status_code
             if self._raise_for_status:
@@ -502,7 +554,7 @@ class Request:
         return self
 
     @staticmethod
-    def _create_request(method: Method, url: str, **kwargs) -> Request:
+    def _create_request(method: Method, url: str, **kwargs) -> httpx.Request:
         """
         Create a request. This is a httpx request wrapper function.
         Args:
@@ -515,7 +567,9 @@ class Request:
         request = httpx.Request(method, url, **kwargs)
         return request
 
-    def _validate_response_data(self, response: ResponseJson) -> ResponseJson:
+    def _validate_response_data(
+        self, response: ResponseJson
+    ) -> Union[BaseModel, ResponseJson | None]:
         """
         Validate response using given validation methods. If there is a validation method and response is not valid,
         validation functions will raise an exception for them.
@@ -531,13 +585,11 @@ class Request:
         try:
             # If a validation model is provided, validate response using the validation model.
             if self._validation_model:
-                validated_response = self._validate_response_by_model(
-                    validated_response
-                )
+                validated_response = self._validate_response_by_model(response)
             # Then, If a validation function is provided, validate response using the validation function.
             if self._validation_function:
                 validated_response = self._validate_response_by_validation_function(
-                    validated_response
+                    response
                 )
         except Exception as exception:
             # If one of the validation methods does not confirm, raised exception will be silently handled.
@@ -546,7 +598,7 @@ class Request:
 
         return validated_response
 
-    def _validate_response_by_model(self, response: ResponseJson) -> ResponseJson:
+    def _validate_response_by_model(self, response: ResponseJson) -> BaseModel:
         """
         Validate response json using the validation model.
         Args:
@@ -554,12 +606,14 @@ class Request:
         Returns:
             ResponseJson: Validated Json object.
         """
-        validated_data = parse_obj_as(self._validation_model, response)
+        validated_data = BaseModel()
+        if self._validation_model:
+            validated_data = parse_obj_as(self._validation_model, response)
         return validated_data
 
     def _validate_response_by_validation_function(
         self, response: ResponseJson
-    ) -> ResponseJson:
+    ) -> ResponseJson | None:
         """
         Validate response json using the validation function.
         Args:
@@ -567,7 +621,11 @@ class Request:
         Returns:
             ResponseJson: Validated Json object.
         """
-        validated_data = self._validation_function(response)
+        validated_data = None
+
+        if self._validation_function:
+            validated_data = self._validation_function(response)
+
         return validated_data
 
     def is_valid(self, raise_exceptions: bool = False) -> bool:
@@ -578,7 +636,7 @@ class Request:
         Returns:
             bool: validity of the data
         """
-        if self.has_exception:
+        if self.has_exception and self._exception:
             if raise_exceptions:
                 raise self._exception
             return False
@@ -603,7 +661,7 @@ class Request:
 
     @property
     def raise_exceptions(self):
-        if self.has_exception:
+        if self.has_exception and self._exception:
             raise self._exception
 
     @property
@@ -612,7 +670,7 @@ class Request:
 
 
 @contextmanager
-def set_directory(path: Path):
+def set_directory(path: Path | str):
     """Context manager that sets the working directory to the given path."""
     origin = Path().absolute()
     try:
@@ -622,10 +680,17 @@ def set_directory(path: Path):
         os.chdir(origin)
 
 
-def strip_invalid_filename_characters(filename: str, max_size: int = 200) -> str:
-    return ("".join([char for char in filename if char.isalnum() or char in "._- "]))[
-        :max_size
-    ]
+def strip_invalid_filename_characters(filename: str, max_bytes: int = 200) -> str:
+    """Strips invalid characters from a filename and ensures that the file_length is less than `max_bytes` bytes."""
+    filename = "".join([char for char in filename if char.isalnum() or char in "._- "])
+    filename_len = len(filename.encode())
+    if filename_len > max_bytes:
+        while filename_len > max_bytes:
+            if len(filename) == 0:
+                break
+            filename = filename[:-1]
+            filename_len = len(filename.encode())
+    return filename
 
 
 def sanitize_value_for_csv(value: str | Number) -> str | Number:
@@ -644,9 +709,7 @@ def sanitize_value_for_csv(value: str | Number) -> str | Number:
     return value
 
 
-def sanitize_list_for_csv(
-    values: List[str | Number] | List[List[str | Number]],
-) -> List[str | Number] | List[List[str | Number]]:
+def sanitize_list_for_csv(values: List[Any]) -> List[Any]:
     """
     Sanitizes a list of values (or a list of list of values) that is being written to a
     CSV file to prevent CSV injection attacks.
@@ -664,29 +727,83 @@ def sanitize_list_for_csv(
 
 def append_unique_suffix(name: str, list_of_names: List[str]):
     """Appends a numerical suffix to `name` so that it does not appear in `list_of_names`."""
-    list_of_names = set(list_of_names)  # for O(1) lookup
-    if name not in list_of_names:
+    set_of_names: set[str] = set(list_of_names)  # for O(1) lookup
+    if name not in set_of_names:
         return name
     else:
         suffix_counter = 1
         new_name = name + f"_{suffix_counter}"
-        while new_name in list_of_names:
+        while new_name in set_of_names:
             suffix_counter += 1
             new_name = name + f"_{suffix_counter}"
         return new_name
 
 
 def validate_url(possible_url: str) -> bool:
+    headers = {"User-Agent": "gradio (https://gradio.app/; team@gradio.app)"}
     try:
-        if requests.get(possible_url).status_code == 200:
-            return True
+        head_request = requests.head(possible_url, headers=headers)
+        if head_request.status_code == 405:
+            return requests.get(possible_url, headers=headers).ok
+        return head_request.ok
     except Exception:
-        pass
-    return False
+        return False
 
 
 def is_update(val):
-    return type(val) is dict and "update" in val.get("__type__", "")
+    return isinstance(val, dict) and "update" in val.get("__type__", "")
+
+
+def get_continuous_fn(fn: Callable, every: float) -> Callable:
+    def continuous_fn(*args):
+        while True:
+            output = fn(*args)
+            yield output
+            time.sleep(every)
+
+    return continuous_fn
+
+
+async def cancel_tasks(task_ids: set[str]):
+    if sys.version_info < (3, 8):
+        return None
+
+    matching_tasks = [
+        task for task in asyncio.all_tasks() if task.get_name() in task_ids
+    ]
+    for task in matching_tasks:
+        task.cancel()
+    await asyncio.gather(*matching_tasks, return_exceptions=True)
+
+
+def set_task_name(task, session_hash: str, fn_index: int, batch: bool):
+    if sys.version_info >= (3, 8) and not (
+        batch
+    ):  # You shouldn't be able to cancel a task if it's part of a batch
+        task.set_name(f"{session_hash}_{fn_index}")
+
+
+def get_cancel_function(
+    dependencies: List[Dict[str, Any]]
+) -> Tuple[Callable, List[int]]:
+    fn_to_comp = {}
+    for dep in dependencies:
+        if Context.root_block:
+            fn_index = next(
+                i for i, d in enumerate(Context.root_block.dependencies) if d == dep
+            )
+            fn_to_comp[fn_index] = [
+                Context.root_block.blocks[o] for o in dep["outputs"]
+            ]
+
+    async def cancel(session_hash: str) -> None:
+        task_ids = set([f"{session_hash}_{fn}" for fn in fn_to_comp])
+        await cancel_tasks(task_ids)
+
+    return (
+        cancel,
+        list(fn_to_comp.keys()),
+    )
 
 
 def check_function_inputs_match(fn: Callable, inputs: List, inputs_as_dict: bool):
@@ -694,17 +811,27 @@ def check_function_inputs_match(fn: Callable, inputs: List, inputs_as_dict: bool
     Checks if the input component set matches the function
     Returns: None if valid, a string error message if mismatch
     """
+
+    def is_special_typed_parameter(name):
+        from gradio.routes import Request
+
+        """Checks if parameter has a type hint designating it as a gr.Request"""
+        return parameter_types.get(name, "") == Request
+
     signature = inspect.signature(fn)
+    parameter_types = typing.get_type_hints(fn) if inspect.isfunction(fn) else {}
     min_args = 0
     max_args = 0
-    for param in signature.parameters.values():
+    infinity = -1
+    for name, param in signature.parameters.items():
         has_default = param.default != param.empty
         if param.kind in [param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD]:
-            if not has_default:
-                min_args += 1
-            max_args += 1
+            if not (is_special_typed_parameter(name)):
+                if not has_default:
+                    min_args += 1
+                max_args += 1
         elif param.kind == param.VAR_POSITIONAL:
-            max_args = "infinity"
+            max_args = infinity
         elif param.kind == param.KEYWORD_ONLY:
             if not has_default:
                 return f"Keyword-only args must have default values for function {fn}"
@@ -717,36 +844,49 @@ def check_function_inputs_match(fn: Callable, inputs: List, inputs_as_dict: bool
         warnings.warn(
             f"Expected at least {min_args} arguments for function {fn}, received {arg_count}."
         )
-    if max_args != "infinity" and arg_count > max_args:
+    if max_args != infinity and arg_count > max_args:
         warnings.warn(
             f"Expected maximum {max_args} arguments for function {fn}, received {arg_count}."
         )
 
 
-def get_cancel_function(
-    dependencies: List[Dict[str, Any]]
-) -> Tuple[Callable, List[int]]:
-    fn_to_comp = {}
-    for dep in dependencies:
-        fn_index = next(
-            i for i, d in enumerate(Context.root_block.dependencies) if d == dep
-        )
-        fn_to_comp[fn_index] = [Context.root_block.blocks[o] for o in dep["outputs"]]
+class TupleNoPrint(tuple):
+    # To remove printing function return in notebook
+    def __repr__(self):
+        return ""
 
-    async def cancel(session_hash: str) -> None:
-        if sys.version_info < (3, 8):
-            return None
+    def __str__(self):
+        return ""
 
-        task_ids = set([f"{session_hash}_{fn}" for fn in fn_to_comp])
 
-        matching_tasks = [
-            task for task in asyncio.all_tasks() if task.get_name() in task_ids
-        ]
-        for task in matching_tasks:
-            task.cancel()
-        await asyncio.gather(*matching_tasks, return_exceptions=True)
-
-    return (
-        cancel,
-        list(fn_to_comp.keys()),
+def tex2svg(formula, *args):
+    FONTSIZE = 20
+    DPI = 300
+    plt.rc("mathtext", fontset="cm")
+    fig = plt.figure(figsize=(0.01, 0.01))
+    fig.text(0, 0, r"${}$".format(formula), fontsize=FONTSIZE)
+    output = BytesIO()
+    fig.savefig(
+        output,
+        dpi=DPI,
+        transparent=True,
+        format="svg",
+        bbox_inches="tight",
+        pad_inches=0.0,
     )
+    plt.close(fig)
+    output.seek(0)
+    xml_code = output.read().decode("utf-8")
+    svg_start = xml_code.index("<svg ")
+    svg_code = xml_code[svg_start:]
+    svg_code = re.sub(r"<metadata>.*<\/metadata>", "", svg_code, flags=re.DOTALL)
+    copy_code = f"<span style='font-size: 0px'>{formula}</span>"
+    return f"{copy_code}{svg_code}"
+
+
+def abspath(path: str | Path) -> Path:
+    """Returns absolute path of a str or Path path, but does not resolve symlinks."""
+    if Path(path).is_symlink():
+        return Path.cwd() / path
+    else:
+        return Path(path).resolve()
