@@ -1,24 +1,25 @@
 import json
 import os
-import pathlib
 import sys
 import textwrap
-import unittest
+import warnings
+from pathlib import Path
 from unittest.mock import MagicMock, patch
 
 import pytest
-import transformers
+from fastapi.testclient import TestClient
 
 import gradio
 import gradio as gr
-from gradio import utils
+from gradio import media_data, utils
+from gradio.exceptions import InvalidApiName
 from gradio.external import (
     TooManyRequestsError,
     cols_to_rows,
-    get_pred_from_ws,
     get_tabular_examples,
     use_websocket,
 )
+from gradio.external_utils import get_pred_from_ws
 
 """
 WARNING: These tests have an external dependency: namely that Hugging Face's
@@ -35,7 +36,7 @@ os.environ["GRADIO_ANALYTICS_ENABLED"] = "False"
 pytestmark = pytest.mark.flaky
 
 
-class TestLoadInterface(unittest.TestCase):
+class TestLoadInterface:
     def test_audio_to_audio(self):
         model_type = "audio-to-audio"
         interface = gr.Interface.load(
@@ -43,52 +44,60 @@ class TestLoadInterface(unittest.TestCase):
             src="models",
             alias=model_type,
         )
-        self.assertEqual(interface.__name__, model_type)
-        self.assertIsInstance(interface.input_components[0], gr.components.Audio)
-        self.assertIsInstance(interface.output_components[0], gr.components.Audio)
+        assert interface.__name__ == model_type
+        assert isinstance(interface.input_components[0], gr.Audio)
+        assert isinstance(interface.output_components[0], gr.Audio)
 
     def test_question_answering(self):
         model_type = "image-classification"
         interface = gr.Blocks.load(
-            name="lysandre/tiny-vit-random", src="models", alias=model_type
+            name="lysandre/tiny-vit-random",
+            src="models",
+            alias=model_type,
         )
-        self.assertEqual(interface.__name__, model_type)
-        self.assertIsInstance(interface.input_components[0], gr.components.Image)
-        self.assertIsInstance(interface.output_components[0], gr.components.Label)
+        assert interface.__name__ == model_type
+        assert isinstance(interface.input_components[0], gr.Image)
+        assert isinstance(interface.output_components[0], gr.Label)
 
     def test_text_generation(self):
         model_type = "text_generation"
-        interface = gr.Interface.load("models/gpt2", alias=model_type)
-        self.assertEqual(interface.__name__, model_type)
-        self.assertIsInstance(interface.input_components[0], gr.components.Textbox)
-        self.assertIsInstance(interface.output_components[0], gr.components.Textbox)
+        interface = gr.Interface.load(
+            "models/gpt2", alias=model_type, description="This is a test description"
+        )
+        assert interface.__name__ == model_type
+        assert isinstance(interface.input_components[0], gr.Textbox)
+        assert isinstance(interface.output_components[0], gr.Textbox)
+        assert any(
+            "This is a test description" in d["props"].get("value", "")
+            for d in interface.get_config_file()["components"]
+        )
 
     def test_summarization(self):
         model_type = "summarization"
         interface = gr.Interface.load(
             "models/facebook/bart-large-cnn", api_key=None, alias=model_type
         )
-        self.assertEqual(interface.__name__, model_type)
-        self.assertIsInstance(interface.input_components[0], gr.components.Textbox)
-        self.assertIsInstance(interface.output_components[0], gr.components.Textbox)
+        assert interface.__name__ == model_type
+        assert isinstance(interface.input_components[0], gr.Textbox)
+        assert isinstance(interface.output_components[0], gr.Textbox)
 
     def test_translation(self):
         model_type = "translation"
         interface = gr.Interface.load(
             "models/facebook/bart-large-cnn", api_key=None, alias=model_type
         )
-        self.assertEqual(interface.__name__, model_type)
-        self.assertIsInstance(interface.input_components[0], gr.components.Textbox)
-        self.assertIsInstance(interface.output_components[0], gr.components.Textbox)
+        assert interface.__name__ == model_type
+        assert isinstance(interface.input_components[0], gr.Textbox)
+        assert isinstance(interface.output_components[0], gr.Textbox)
 
     def test_text2text_generation(self):
         model_type = "text2text-generation"
         interface = gr.Interface.load(
             "models/sshleifer/tiny-mbart", api_key=None, alias=model_type
         )
-        self.assertEqual(interface.__name__, model_type)
-        self.assertIsInstance(interface.input_components[0], gr.components.Textbox)
-        self.assertIsInstance(interface.output_components[0], gr.components.Textbox)
+        assert interface.__name__ == model_type
+        assert isinstance(interface.input_components[0], gr.Textbox)
+        assert isinstance(interface.output_components[0], gr.Textbox)
 
     def test_text_classification(self):
         model_type = "text-classification"
@@ -97,47 +106,47 @@ class TestLoadInterface(unittest.TestCase):
             api_key=None,
             alias=model_type,
         )
-        self.assertEqual(interface.__name__, model_type)
-        self.assertIsInstance(interface.input_components[0], gr.components.Textbox)
-        self.assertIsInstance(interface.output_components[0], gr.components.Label)
+        assert interface.__name__ == model_type
+        assert isinstance(interface.input_components[0], gr.Textbox)
+        assert isinstance(interface.output_components[0], gr.Label)
 
     def test_fill_mask(self):
         model_type = "fill-mask"
         interface = gr.Interface.load(
             "models/bert-base-uncased", api_key=None, alias=model_type
         )
-        self.assertEqual(interface.__name__, model_type)
-        self.assertIsInstance(interface.input_components[0], gr.components.Textbox)
-        self.assertIsInstance(interface.output_components[0], gr.components.Label)
+        assert interface.__name__ == model_type
+        assert isinstance(interface.input_components[0], gr.Textbox)
+        assert isinstance(interface.output_components[0], gr.Label)
 
     def test_zero_shot_classification(self):
         model_type = "zero-shot-classification"
         interface = gr.Interface.load(
             "models/facebook/bart-large-mnli", api_key=None, alias=model_type
         )
-        self.assertEqual(interface.__name__, model_type)
-        self.assertIsInstance(interface.input_components[0], gr.components.Textbox)
-        self.assertIsInstance(interface.input_components[1], gr.components.Textbox)
-        self.assertIsInstance(interface.input_components[2], gr.components.Checkbox)
-        self.assertIsInstance(interface.output_components[0], gr.components.Label)
+        assert interface.__name__ == model_type
+        assert isinstance(interface.input_components[0], gr.Textbox)
+        assert isinstance(interface.input_components[1], gr.Textbox)
+        assert isinstance(interface.input_components[2], gr.Checkbox)
+        assert isinstance(interface.output_components[0], gr.Label)
 
     def test_automatic_speech_recognition(self):
         model_type = "automatic-speech-recognition"
         interface = gr.Interface.load(
             "models/facebook/wav2vec2-base-960h", api_key=None, alias=model_type
         )
-        self.assertEqual(interface.__name__, model_type)
-        self.assertIsInstance(interface.input_components[0], gr.components.Audio)
-        self.assertIsInstance(interface.output_components[0], gr.components.Textbox)
+        assert interface.__name__ == model_type
+        assert isinstance(interface.input_components[0], gr.Audio)
+        assert isinstance(interface.output_components[0], gr.Textbox)
 
     def test_image_classification(self):
         model_type = "image-classification"
         interface = gr.Interface.load(
             "models/google/vit-base-patch16-224", api_key=None, alias=model_type
         )
-        self.assertEqual(interface.__name__, model_type)
-        self.assertIsInstance(interface.input_components[0], gr.components.Image)
-        self.assertIsInstance(interface.output_components[0], gr.components.Label)
+        assert interface.__name__ == model_type
+        assert isinstance(interface.input_components[0], gr.Image)
+        assert isinstance(interface.output_components[0], gr.Label)
 
     def test_feature_extraction(self):
         model_type = "feature-extraction"
@@ -146,9 +155,9 @@ class TestLoadInterface(unittest.TestCase):
             api_key=None,
             alias=model_type,
         )
-        self.assertEqual(interface.__name__, model_type)
-        self.assertIsInstance(interface.input_components[0], gr.components.Textbox)
-        self.assertIsInstance(interface.output_components[0], gr.components.Dataframe)
+        assert interface.__name__ == model_type
+        assert isinstance(interface.input_components[0], gr.Textbox)
+        assert isinstance(interface.output_components[0], gr.Dataframe)
 
     def test_sentence_similarity(self):
         model_type = "text-to-speech"
@@ -157,9 +166,9 @@ class TestLoadInterface(unittest.TestCase):
             api_key=None,
             alias=model_type,
         )
-        self.assertEqual(interface.__name__, model_type)
-        self.assertIsInstance(interface.input_components[0], gr.components.Textbox)
-        self.assertIsInstance(interface.output_components[0], gr.components.Audio)
+        assert interface.__name__ == model_type
+        assert isinstance(interface.input_components[0], gr.Textbox)
+        assert isinstance(interface.output_components[0], gr.Audio)
 
     def test_text_to_speech(self):
         model_type = "text-to-speech"
@@ -168,23 +177,24 @@ class TestLoadInterface(unittest.TestCase):
             api_key=None,
             alias=model_type,
         )
-        self.assertEqual(interface.__name__, model_type)
-        self.assertIsInstance(interface.input_components[0], gr.components.Textbox)
-        self.assertIsInstance(interface.output_components[0], gr.components.Audio)
+        assert interface.__name__ == model_type
+        assert isinstance(interface.input_components[0], gr.Textbox)
+        assert isinstance(interface.output_components[0], gr.Audio)
 
     def test_text_to_image(self):
         model_type = "text-to-image"
         interface = gr.Interface.load(
             "models/osanseviero/BigGAN-deep-128", api_key=None, alias=model_type
         )
-        self.assertEqual(interface.__name__, model_type)
-        self.assertIsInstance(interface.input_components[0], gr.components.Textbox)
-        self.assertIsInstance(interface.output_components[0], gr.components.Image)
+        assert interface.__name__ == model_type
+        assert isinstance(interface.input_components[0], gr.Textbox)
+        assert isinstance(interface.output_components[0], gr.Image)
 
     def test_english_to_spanish(self):
-        interface = gr.Interface.load("spaces/abidlabs/english_to_spanish")
-        self.assertIsInstance(interface.input_components[0], gr.components.Textbox)
-        self.assertIsInstance(interface.output_components[0], gr.components.Textbox)
+        with pytest.warns(UserWarning):
+            io = gr.Interface.load("spaces/abidlabs/english_to_spanish", title="hi")
+        assert isinstance(io.input_components[0], gr.Textbox)
+        assert isinstance(io.output_components[0], gr.Textbox)
 
     def test_sentiment_model(self):
         io = gr.Interface.load("models/distilbert-base-uncased-finetuned-sst-2-english")
@@ -206,7 +216,7 @@ class TestLoadInterface(unittest.TestCase):
         io = gr.Blocks.load(name="models/t5-base")
         try:
             output = io("My name is Sarah and I live in London")
-            self.assertEqual(output, "Mein Name ist Sarah und ich lebe in London")
+            assert output == "Mein Name ist Sarah und ich lebe in London"
         except TooManyRequestsError:
             pass
 
@@ -218,49 +228,120 @@ class TestLoadInterface(unittest.TestCase):
         except TooManyRequestsError:
             pass
 
+    def test_image_to_text(self):
+        io = gr.Interface.load("models/nlpconnect/vit-gpt2-image-captioning")
+        try:
+            output = io("gradio/test_data/lion.jpg")
+            assert isinstance(output, str)
+        except TooManyRequestsError:
+            pass
+
+    def test_conversational(self):
+        io = gr.Interface.load("models/microsoft/DialoGPT-medium")
+        app, _, _ = io.launch(prevent_thread_lock=True)
+        client = TestClient(app)
+        assert app.state_holder == {}
+        response = client.post(
+            "/api/predict/",
+            json={"session_hash": "foo", "data": ["Hi!", None], "fn_index": 0},
+        )
+        output = response.json()
+        assert isinstance(output["data"], list)
+        assert isinstance(output["data"][0], list)
+        assert isinstance(app.state_holder["foo"], dict)
+
     def test_speech_recognition_model(self):
         io = gr.Interface.load("models/facebook/wav2vec2-base-960h")
         try:
             output = io("gradio/test_data/test_audio.wav")
-            self.assertIsNotNone(output)
+            assert output is not None
         except TooManyRequestsError:
             pass
 
+        app, _, _ = io.launch(prevent_thread_lock=True, show_error=True)
+        client = TestClient(app)
+        resp = client.post(
+            "api/predict",
+            json={"fn_index": 0, "data": [media_data.BASE64_AUDIO], "name": "sample"},
+        )
+        try:
+            if resp.status_code != 200:
+                warnings.warn("Request for speech recognition model failed!")
+                if (
+                    "Could not complete request to HuggingFace API"
+                    in resp.json()["error"]
+                ):
+                    pass
+                else:
+                    assert False
+            else:
+                assert resp.json()["data"] is not None
+        finally:
+            io.close()
+
     def test_text_to_image_model(self):
         io = gr.Interface.load("models/osanseviero/BigGAN-deep-128")
         try:
             filename = io("chest")
-            self.assertTrue(filename.endswith(".jpg") or filename.endswith(".jpeg"))
+            assert filename.endswith(".jpg") or filename.endswith(".jpeg")
+        except TooManyRequestsError:
+            pass
+
+    def test_private_space(self):
+        api_key = "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"  # Intentionally revealing this key for testing purposes
+        io = gr.Interface.load(
+            "spaces/gradio-tests/not-actually-private-space", api_key=api_key
+        )
+        try:
+            output = io("abc")
+            assert output == "abc"
         except TooManyRequestsError:
             pass
 
 
-class TestLoadFromPipeline(unittest.TestCase):
-    def test_text_to_text_model_from_pipeline(self):
-        pipe = transformers.pipeline(model="sshleifer/bart-tiny-random")
-        output = pipe("My name is Sylvain and I work at Hugging Face in Brooklyn")
-        self.assertIsNotNone(output)
-
-
 class TestLoadInterfaceWithExamples:
     def test_interface_load_examples(self, tmp_path):
-        test_file_dir = pathlib.Path(pathlib.Path(__file__).parent, "test_files")
-        with patch("gradio.examples.CACHED_FOLDER", tmp_path):
+        test_file_dir = Path(Path(__file__).parent, "test_files")
+        with patch("gradio.helpers.CACHED_FOLDER", tmp_path):
             gr.Interface.load(
                 name="models/google/vit-base-patch16-224",
-                examples=[pathlib.Path(test_file_dir, "cheetah1.jpg")],
+                examples=[Path(test_file_dir, "cheetah1.jpg")],
                 cache_examples=False,
             )
 
     def test_interface_load_cache_examples(self, tmp_path):
-        test_file_dir = pathlib.Path(pathlib.Path(__file__).parent, "test_files")
-        with patch("gradio.examples.CACHED_FOLDER", tmp_path):
+        test_file_dir = Path(Path(__file__).parent, "test_files")
+        with patch("gradio.helpers.CACHED_FOLDER", tmp_path):
             gr.Interface.load(
                 name="models/google/vit-base-patch16-224",
-                examples=[pathlib.Path(test_file_dir, "cheetah1.jpg")],
+                examples=[Path(test_file_dir, "cheetah1.jpg")],
                 cache_examples=True,
             )
 
+    def test_root_url(self):
+        demo = gr.Interface.load("spaces/gradio/test-loading-examples")
+        assert all(
+            [
+                c["props"]["root_url"]
+                == "https://gradio-test-loading-examples.hf.space/"
+                for c in demo.get_config_file()["components"]
+            ]
+        )
+
+    def test_root_url_deserialization(self):
+        demo = gr.Interface.load("spaces/gradio/simple_gallery")
+        path_to_files = demo("test")
+        assert (Path(path_to_files) / "captions.json").exists()
+
+    def test_interface_with_examples(self):
+        # This demo has the "fake_event" correctly removed
+        demo = gr.Interface.load("spaces/freddyaboulton/calculator")
+        assert demo(2, "add", 3) == 5
+
+        # This demo still has the "fake_event". both should work
+        demo = gr.Interface.load("spaces/abidlabs/test-calculator-2")
+        assert demo(2, "add", 4) == 6
+
 
 def test_get_tabular_examples_replaces_nan_with_str_nan():
     readme = """
@@ -330,6 +411,20 @@ def check_dataset(config, readme_examples):
         ]
 
 
+def test_load_blocks_with_default_values():
+    io = gr.Interface.load("spaces/abidlabs/min-dalle")
+    assert isinstance(io.get_config_file()["components"][0]["props"]["value"], list)
+
+    io = gr.Interface.load("spaces/abidlabs/min-dalle-later")
+    assert isinstance(io.get_config_file()["components"][0]["props"]["value"], list)
+
+    io = gr.Interface.load("spaces/freddyaboulton/dataframe_load")
+    assert io.get_config_file()["components"][0]["props"]["value"] == {
+        "headers": ["a", "b"],
+        "data": [[1, 4], [2, 5], [3, 6]],
+    }
+
+
 @pytest.mark.parametrize(
     "hypothetical_readme",
     [
@@ -406,13 +501,26 @@ async def test_get_pred_from_ws_raises_if_queue_full():
     reason="Mocks of async context manager don't work for 3.7",
 )
 def test_respect_queue_when_load_from_config():
-    with unittest.mock.patch("websockets.connect"):
-        with unittest.mock.patch(
-            "gradio.external.get_pred_from_ws", return_value={"data": ["foo"]}
+    with patch("websockets.connect"):
+        with patch(
+            "gradio.external_utils.get_pred_from_ws", return_value={"data": ["foo"]}
         ):
             interface = gr.Interface.load("spaces/freddyaboulton/saymyname")
             assert interface("bob") == "foo"
 
 
-if __name__ == "__main__":
-    unittest.main()
+def test_raise_value_error_when_api_name_invalid():
+    with pytest.raises(InvalidApiName):
+        demo = gr.Blocks.load(name="spaces/gradio/hello_world")
+        demo("freddy", api_name="route does not exist")
+
+
+def test_use_api_name_in_call_method():
+    # Interface
+    demo = gr.Blocks.load(name="spaces/gradio/hello_world")
+    assert demo("freddy", api_name="predict") == "Hello freddy!"
+
+    # Blocks demo with multiple functions
+    app = gr.Blocks.load(name="spaces/gradio/multiple-api-name-test")
+    assert app(15, api_name="minus_one") == 14
+    assert app(4, api_name="double") == 8
