@@ -2,9 +2,10 @@
 import json
 import os
 import sys
-import unittest
 from unittest.mock import patch
 
+import numpy as np
+import pandas as pd
 import pytest
 import starlette.routing
 import websockets
@@ -12,39 +13,53 @@ from fastapi import FastAPI
 from fastapi.testclient import TestClient
 
 import gradio as gr
-from gradio import Blocks, Button, Interface, Number, Textbox, close_all, routes
+from gradio import (
+    Blocks,
+    Button,
+    Interface,
+    Number,
+    Textbox,
+    close_all,
+    media_data,
+    routes,
+)
 
 os.environ["GRADIO_ANALYTICS_ENABLED"] = "False"
 
 
-class TestRoutes(unittest.TestCase):
-    def setUp(self) -> None:
-        self.io = Interface(lambda x: x + x, "text", "text")
-        self.app, _, _ = self.io.launch(prevent_thread_lock=True)
-        self.client = TestClient(self.app)
+@pytest.fixture()
+def test_client():
+    io = Interface(lambda x: x + x, "text", "text")
+    app, _, _ = io.launch(prevent_thread_lock=True)
+    test_client = TestClient(app)
+    yield test_client
+    io.close()
+    close_all()
 
-    def test_get_main_route(self):
-        response = self.client.get("/")
-        self.assertEqual(response.status_code, 200)
 
-    def test_static_files_served_safely(self):
+class TestRoutes:
+    def test_get_main_route(self, test_client):
+        response = test_client.get("/")
+        assert response.status_code == 200
+
+    def test_static_files_served_safely(self, test_client):
         # Make sure things outside the static folder are not accessible
-        response = self.client.get(r"/static/..%2findex.html")
-        self.assertEqual(response.status_code, 404)
-        response = self.client.get(r"/static/..%2f..%2fapi_docs.html")
-        self.assertEqual(response.status_code, 404)
+        response = test_client.get(r"/static/..%2findex.html")
+        assert response.status_code == 404
+        response = test_client.get(r"/static/..%2f..%2fapi_docs.html")
+        assert response.status_code == 404
 
-    def test_get_config_route(self):
-        response = self.client.get("/config/")
-        self.assertEqual(response.status_code, 200)
+    def test_get_config_route(self, test_client):
+        response = test_client.get("/config/")
+        assert response.status_code == 200
 
-    def test_predict_route(self):
-        response = self.client.post(
+    def test_predict_route(self, test_client):
+        response = test_client.post(
             "/api/predict/", json={"data": ["test"], "fn_index": 0}
         )
-        self.assertEqual(response.status_code, 200)
+        assert response.status_code == 200
         output = dict(response.json())
-        self.assertEqual(output["data"], ["testtest"])
+        assert output["data"] == ["testtest"]
 
     def test_named_predict_route(self):
         with Blocks() as demo:
@@ -109,11 +124,11 @@ class TestRoutes(unittest.TestCase):
         output = dict(response.json())
         assert output["data"] == ["test2"]
 
-    def test_predict_route_without_fn_index(self):
-        response = self.client.post("/api/predict/", json={"data": ["test"]})
-        self.assertEqual(response.status_code, 200)
+    def test_predict_route_without_fn_index(self, test_client):
+        response = test_client.post("/api/predict/", json={"data": ["test"]})
+        assert response.status_code == 200
         output = dict(response.json())
-        self.assertEqual(output["data"], ["testtest"])
+        assert output["data"] == ["testtest"]
 
     def test_predict_route_batching(self):
         def batch_fn(x):
@@ -132,7 +147,7 @@ class TestRoutes(unittest.TestCase):
         client = TestClient(app)
         response = client.post("/api/pred/", json={"data": ["test"]})
         output = dict(response.json())
-        self.assertEqual(output["data"], ["Hello test"])
+        assert output["data"] == ["Hello test"]
 
         app, _, _ = demo.launch(prevent_thread_lock=True)
         client = TestClient(app)
@@ -140,7 +155,7 @@ class TestRoutes(unittest.TestCase):
             "/api/pred/", json={"data": [["test", "test2"]], "batched": True}
         )
         output = dict(response.json())
-        self.assertEqual(output["data"], [["Hello test", "Hello test2"]])
+        assert output["data"] == [["Hello test", "Hello test2"]]
 
     def test_state(self):
         def predict(input, history):
@@ -157,17 +172,69 @@ class TestRoutes(unittest.TestCase):
             json={"data": ["test", None], "fn_index": 0, "session_hash": "_"},
         )
         output = dict(response.json())
-        self.assertEqual(output["data"], ["test", None])
+        assert output["data"] == ["test", None]
         response = client.post(
             "/api/predict/",
             json={"data": ["test", None], "fn_index": 0, "session_hash": "_"},
         )
         output = dict(response.json())
-        self.assertEqual(output["data"], ["testtest", None])
+        assert output["data"] == ["testtest", None]
 
-    def tearDown(self) -> None:
-        self.io.close()
-        close_all()
+    def test_get_file_created_by_app(self):
+        app, _, _ = gr.Interface(lambda s: s.name, gr.File(), gr.File()).launch(
+            prevent_thread_lock=True
+        )
+        client = TestClient(app)
+        response = client.post(
+            "/api/predict/",
+            json={
+                "data": [
+                    {
+                        "data": media_data.BASE64_IMAGE,
+                        "name": "bus.png",
+                        "size": len(media_data.BASE64_IMAGE),
+                    }
+                ],
+                "fn_index": 0,
+                "session_hash": "_",
+            },
+        ).json()
+        created_file = response["data"][0]["name"]
+        file_response = client.get(f"/file={created_file}")
+        assert file_response.is_success
+
+        backwards_compatible_file_response = client.get(f"/file/{created_file}")
+        assert backwards_compatible_file_response.is_success
+
+        file_response_with_full_range = client.get(
+            f"/file={created_file}", headers={"Range": "bytes=0-"}
+        )
+        assert file_response_with_full_range.is_success
+        assert file_response.text == file_response_with_full_range.text
+
+        file_response_with_partial_range = client.get(
+            f"/file={created_file}", headers={"Range": "bytes=0-10"}
+        )
+        assert file_response_with_partial_range.is_success
+        assert len(file_response_with_partial_range.text) == 11
+
+    def test_mount_gradio_app(self):
+        app = FastAPI()
+
+        demo = gr.Interface(
+            lambda s: f"Hello from ps, {s}!", "textbox", "textbox"
+        ).queue()
+        demo1 = gr.Interface(
+            lambda s: f"Hello from py, {s}!", "textbox", "textbox"
+        ).queue()
+
+        app = gr.mount_gradio_app(app, demo, path="/ps")
+        app = gr.mount_gradio_app(app, demo1, path="/py")
+
+        # Use context manager to trigger start up events
+        with TestClient(app) as client:
+            assert client.get("/ps").is_success
+            assert client.get("/py").is_success
 
 
 class TestGeneratorRoutes:
@@ -227,29 +294,26 @@ class TestApp:
         assert isinstance(app, FastAPI)
 
 
-class TestAuthenticatedRoutes(unittest.TestCase):
-    def setUp(self) -> None:
-        self.io = Interface(lambda x: x, "text", "text")
-        self.app, _, _ = self.io.launch(
+class TestAuthenticatedRoutes:
+    def test_post_login(self):
+        io = Interface(lambda x: x, "text", "text")
+        app, _, _ = io.launch(
             auth=("test", "correct_password"),
             prevent_thread_lock=True,
             enable_queue=False,
         )
-        self.client = TestClient(self.app)
+        client = TestClient(app)
 
-    def test_post_login(self):
-        response = self.client.post(
-            "/login", data=dict(username="test", password="correct_password")
+        response = client.post(
+            "/login",
+            data=dict(username="test", password="correct_password"),
         )
-        self.assertEqual(response.status_code, 302)
-        response = self.client.post(
-            "/login", data=dict(username="test", password="incorrect_password")
+        assert response.status_code == 200
+        response = client.post(
+            "/login",
+            data=dict(username="test", password="incorrect_password"),
         )
-        self.assertEqual(response.status_code, 400)
-
-    def tearDown(self) -> None:
-        self.io.close()
-        close_all()
+        assert response.status_code == 400
 
 
 class TestQueueRoutes:
@@ -285,8 +349,8 @@ class TestQueueRoutes:
                 "http://127.0.0.1:7861/gradio/gradio/gradio/",
             ),
             (
-                "wss://huggingface.co.tech/path/queue/join",
-                "https://huggingface.co.tech/path/",
+                "wss://gradio-titanic-survival.hf.space/queue/join",
+                "https://gradio-titanic-survival.hf.space/",
             ),
         ],
     )
@@ -312,6 +376,23 @@ class TestDevMode:
         assert not gradio_fast_api.app.blocks.dev_mode
 
 
+class TestPassingRequest:
+    def test_request_included_with_regular_function(self):
+        def identity(name, request: gr.Request):
+            assert isinstance(request.client.host, str)
+            return name
+
+        app, _, _ = gr.Interface(identity, "textbox", "textbox").launch(
+            prevent_thread_lock=True,
+        )
+        client = TestClient(app)
+
+        response = client.post("/api/predict/", json={"data": ["test"]})
+        assert response.status_code == 200
+        output = dict(response.json())
+        assert output["data"] == ["test"]
+
+
 def test_predict_route_is_blocked_if_api_open_false():
     io = Interface(lambda x: x, "text", "text", examples=[["freddy"]]).queue(
         api_open=False
@@ -384,5 +465,22 @@ def test_show_api_queue_not_enabled():
     assert not io.show_api
 
 
-if __name__ == "__main__":
-    unittest.main()
+def test_orjson_serialization():
+    df = pd.DataFrame(
+        {
+            "date_1": pd.date_range("2021-01-01", periods=2),
+            "date_2": pd.date_range("2022-02-15", periods=2).strftime("%B %d, %Y, %r"),
+            "number": np.array([0.2233, 0.57281]),
+            "number_2": np.array([84, 23]).astype(np.int64),
+            "bool": [True, False],
+            "markdown": ["# Hello", "# Goodbye"],
+        }
+    )
+
+    with gr.Blocks() as demo:
+        gr.DataFrame(df)
+    app, _, _ = demo.launch(prevent_thread_lock=True)
+    test_client = TestClient(app)
+    response = test_client.get("/")
+    assert response.status_code == 200
+    demo.close()
