@@ -2,7 +2,6 @@ import copy
 import ipaddress
 import json
 import os
-import unittest
 import unittest.mock as mock
 import warnings
 
@@ -13,13 +12,15 @@ from httpx import AsyncClient, Response
 from pydantic import BaseModel
 from typing_extensions import Literal
 
+from gradio.context import Context
 from gradio.test_data.blocks_configs import (
     XRAY_CONFIG,
     XRAY_CONFIG_DIFF_IDS,
     XRAY_CONFIG_WITH_MISTAKE,
 )
 from gradio.utils import (
-    Request,
+    AsyncRequest,
+    abspath,
     append_unique_suffix,
     assert_configs_are_equivalent_besides_ids,
     colab_check,
@@ -28,17 +29,20 @@ from gradio.utils import (
     format_ner_list,
     get_local_ip_address,
     ipython_check,
+    kaggle_check,
     launch_analytics,
     readme_to_html,
     sanitize_list_for_csv,
     sanitize_value_for_csv,
+    strip_invalid_filename_characters,
+    validate_url,
     version_check,
 )
 
 os.environ["GRADIO_ANALYTICS_ENABLED"] = "False"
 
 
-class TestUtils(unittest.TestCase):
+class TestUtils:
     @mock.patch("requests.get")
     def test_should_warn_with_unable_to_parse(self, mock_get):
         mock_get.side_effect = json.decoder.JSONDecodeError("Expecting value", "", 0)
@@ -46,8 +50,9 @@ class TestUtils(unittest.TestCase):
         with warnings.catch_warnings(record=True) as w:
             warnings.simplefilter("always")
             version_check()
-            self.assertEqual(
-                str(w[-1].message), "unable to parse version details from package URL."
+            assert (
+                str(w[-1].message)
+                == "unable to parse version details from package URL."
             )
 
     @mock.patch("requests.Response.json")
@@ -57,19 +62,17 @@ class TestUtils(unittest.TestCase):
         with warnings.catch_warnings(record=True) as w:
             warnings.simplefilter("always")
             version_check()
-            self.assertEqual(
-                str(w[-1].message), "package URL does not contain version info."
-            )
+            assert str(w[-1].message) == "package URL does not contain version info."
 
     @mock.patch("requests.post")
     def test_error_analytics_doesnt_crash_on_connection_error(self, mock_post):
         mock_post.side_effect = requests.ConnectionError()
-        error_analytics("placeholder", "placeholder")
+        error_analytics("placeholder")
         mock_post.assert_called()
 
     @mock.patch("requests.post")
     def test_error_analytics_successful(self, mock_post):
-        error_analytics("placeholder", "placeholder")
+        error_analytics("placeholder")
         mock_post.assert_called()
 
     @mock.patch("requests.post")
@@ -101,37 +104,60 @@ class TestUtils(unittest.TestCase):
     def test_readme_to_html_correct_parse(self):
         readme_to_html("https://github.com/gradio-app/gradio/blob/master/README.md")
 
+    def test_kaggle_check_false(self):
+        assert not kaggle_check()
 
-class TestIPAddress(unittest.TestCase):
+    def test_kaggle_check_true_when_run_type_set(self):
+        with mock.patch.dict(
+            os.environ, {"KAGGLE_KERNEL_RUN_TYPE": "Interactive"}, clear=True
+        ):
+            assert kaggle_check()
+
+    def test_kaggle_check_true_when_both_set(self):
+        with mock.patch.dict(
+            os.environ,
+            {"KAGGLE_KERNEL_RUN_TYPE": "Interactive", "GFOOTBALL_DATA_DIR": "./"},
+            clear=True,
+        ):
+            assert kaggle_check()
+
+    def test_kaggle_check_false_when_neither_set(self):
+        with mock.patch.dict(
+            os.environ,
+            {"KAGGLE_KERNEL_RUN_TYPE": "", "GFOOTBALL_DATA_DIR": ""},
+            clear=True,
+        ):
+            assert not kaggle_check()
+
+
+class TestIPAddress:
+    @pytest.mark.flaky
     def test_get_ip(self):
+        Context.ip_address = None
         ip = get_local_ip_address()
         if ip == "No internet connection":
             return
-        try:  # check whether ip is valid
-            ipaddress.ip_address(ip)
-        except ValueError:
-            self.fail("Invalid IP address")
+        ipaddress.ip_address(ip)
 
     @mock.patch("requests.get")
     def test_get_ip_without_internet(self, mock_get):
+        Context.ip_address = None
         mock_get.side_effect = requests.ConnectionError()
         ip = get_local_ip_address()
-        self.assertEqual(ip, "No internet connection")
+        assert ip == "No internet connection"
 
 
-class TestAssertConfigsEquivalent(unittest.TestCase):
+class TestAssertConfigsEquivalent:
     def test_same_configs(self):
-        self.assertTrue(
-            assert_configs_are_equivalent_besides_ids(XRAY_CONFIG, XRAY_CONFIG)
-        )
+        assert assert_configs_are_equivalent_besides_ids(XRAY_CONFIG, XRAY_CONFIG)
 
     def test_equivalent_configs(self):
-        self.assertTrue(
-            assert_configs_are_equivalent_besides_ids(XRAY_CONFIG, XRAY_CONFIG_DIFF_IDS)
+        assert assert_configs_are_equivalent_besides_ids(
+            XRAY_CONFIG, XRAY_CONFIG_DIFF_IDS
         )
 
     def test_different_configs(self):
-        with self.assertRaises(AssertionError):
+        with pytest.raises(AssertionError):
             assert_configs_are_equivalent_besides_ids(
                 XRAY_CONFIG_WITH_MISTAKE, XRAY_CONFIG
             )
@@ -207,11 +233,11 @@ class TestAssertConfigsEquivalent(unittest.TestCase):
 
         config2 = copy.deepcopy(config1)
         config2["dependencies"][0]["documentation"] = None
-        with self.assertRaises(AssertionError):
+        with pytest.raises(AssertionError):
             assert_configs_are_equivalent_besides_ids(config1, config2)
 
 
-class TestFormatNERList(unittest.TestCase):
+class TestFormatNERList:
     def test_format_ner_list_standard(self):
         string = "Wolfgang lives in Berlin"
         groups = [
@@ -225,16 +251,16 @@ class TestFormatNERList(unittest.TestCase):
             ("Berlin", "LOC"),
             ("", None),
         ]
-        self.assertEqual(format_ner_list(string, groups), result)
+        assert format_ner_list(string, groups) == result
 
     def test_format_ner_list_empty(self):
         string = "I live in a city"
         groups = []
         result = [("I live in a city", None)]
-        self.assertEqual(format_ner_list(string, groups), result)
+        assert format_ner_list(string, groups) == result
 
 
-class TestDeleteNone(unittest.TestCase):
+class TestDeleteNone:
     """Credit: https://stackoverflow.com/questions/33797126/proper-way-to-remove-keys-in-dictionary-with-none-values-in-python"""
 
     def test_delete_none(self):
@@ -250,7 +276,7 @@ class TestDeleteNone(unittest.TestCase):
             },
         }
         truth = {"a": 12, "b": 34, "k": {"d": 34, "m": [{"k": 23}, [1, 2, 3], {1, 2}]}}
-        self.assertEqual(delete_none(input), truth)
+        assert delete_none(input) == truth
 
 
 @pytest_asyncio.fixture(scope="function", autouse=True)
@@ -259,15 +285,15 @@ async def client():
     A fixture to mock the async client object.
     """
     async with AsyncClient() as mock_client:
-        with mock.patch("gradio.utils.Request.client", mock_client):
+        with mock.patch("gradio.utils.AsyncRequest.client", mock_client):
             yield
 
 
 class TestRequest:
     @pytest.mark.asyncio
     async def test_get(self):
-        client_response: Request = await Request(
-            method=Request.Method.GET,
+        client_response: AsyncRequest = await AsyncRequest(
+            method=AsyncRequest.Method.GET,
             url="http://headers.jsontest.com/",
         )
         validated_data = client_response.get_validated_data()
@@ -276,8 +302,8 @@ class TestRequest:
 
     @pytest.mark.asyncio
     async def test_post(self):
-        client_response: Request = await Request(
-            method=Request.Method.POST,
+        client_response: AsyncRequest = await AsyncRequest(
+            method=AsyncRequest.Method.POST,
             url="https://reqres.in/api/users",
             json={"name": "morpheus", "job": "leader"},
         )
@@ -294,8 +320,8 @@ class TestRequest:
             id: str
             createdAt: str
 
-        client_response: Request = await Request(
-            method=Request.Method.POST,
+        client_response: AsyncRequest = await AsyncRequest(
+            method=AsyncRequest.Method.POST,
             url="https://reqres.in/api/users",
             json={"name": "morpheus", "job": "leader"},
             validation_model=TestModel,
@@ -308,8 +334,8 @@ class TestRequest:
             name: Literal["John"] = "John"
             job: str
 
-        client_response: Request = await Request(
-            method=Request.Method.POST,
+        client_response: AsyncRequest = await AsyncRequest(
+            method=AsyncRequest.Method.POST,
             url="https://reqres.in/api/users",
             json={"name": "morpheus", "job": "leader"},
             validation_model=TestModel,
@@ -333,8 +359,8 @@ async def test_get(respx_mock):
         make_mock_response({"Host": "headers.jsontest.com"})
     )
 
-    client_response: Request = await Request(
-        method=Request.Method.GET,
+    client_response: AsyncRequest = await AsyncRequest(
+        method=AsyncRequest.Method.GET,
         url=MOCK_REQUEST_URL,
     )
     validated_data = client_response.get_validated_data()
@@ -348,8 +374,8 @@ async def test_post(respx_mock):
     payload = {"name": "morpheus", "job": "leader"}
     respx_mock.post(MOCK_REQUEST_URL).mock(make_mock_response(payload))
 
-    client_response: Request = await Request(
-        method=Request.Method.POST,
+    client_response: AsyncRequest = await AsyncRequest(
+        method=AsyncRequest.Method.POST,
         url=MOCK_REQUEST_URL,
         json=payload,
     )
@@ -378,8 +404,8 @@ async def test_validate_with_model(respx_mock):
         id: str
         createdAt: str
 
-    client_response: Request = await Request(
-        method=Request.Method.POST,
+    client_response: AsyncRequest = await AsyncRequest(
+        method=AsyncRequest.Method.POST,
         url=MOCK_REQUEST_URL,
         json={"name": "morpheus", "job": "leader"},
         validation_model=TestModel,
@@ -390,14 +416,14 @@ async def test_validate_with_model(respx_mock):
 @pytest.mark.asyncio
 async def test_validate_and_fail_with_model(respx_mock):
     class TestModel(BaseModel):
-        name: Literal[str] = "John"
+        name: Literal["John"]
         job: str
 
     payload = {"name": "morpheus", "job": "leader"}
     respx_mock.post(MOCK_REQUEST_URL).mock(make_mock_response(payload))
 
-    client_response: Request = await Request(
-        method=Request.Method.POST,
+    client_response: AsyncRequest = await AsyncRequest(
+        method=AsyncRequest.Method.POST,
         url=MOCK_REQUEST_URL,
         json=payload,
         validation_model=TestModel,
@@ -408,7 +434,7 @@ async def test_validate_and_fail_with_model(respx_mock):
     assert isinstance(client_response.exception, Exception)
 
 
-@mock.patch("gradio.utils.Request._validate_response_data")
+@mock.patch("gradio.utils.AsyncRequest._validate_response_data")
 @pytest.mark.asyncio
 async def test_exception_type(validate_response_data, respx_mock):
     class ResponseValidationException(Exception):
@@ -418,8 +444,8 @@ async def test_exception_type(validate_response_data, respx_mock):
 
     respx_mock.get(MOCK_REQUEST_URL).mock(Response(201))
 
-    client_response: Request = await Request(
-        method=Request.Method.GET,
+    client_response: AsyncRequest = await AsyncRequest(
+        method=AsyncRequest.Method.GET,
         url=MOCK_REQUEST_URL,
         exception_type=ResponseValidationException,
     )
@@ -438,8 +464,8 @@ async def test_validate_with_function(respx_mock):
             return response
         raise Exception
 
-    client_response: Request = await Request(
-        method=Request.Method.POST,
+    client_response: AsyncRequest = await AsyncRequest(
+        method=AsyncRequest.Method.POST,
         url=MOCK_REQUEST_URL,
         json={"name": "morpheus", "job": "leader"},
         validation_function=has_name,
@@ -460,8 +486,8 @@ async def test_validate_and_fail_with_function(respx_mock):
 
     respx_mock.post(MOCK_REQUEST_URL).mock(make_mock_response({"name": "morpheus"}))
 
-    client_response: Request = await Request(
-        method=Request.Method.POST,
+    client_response: AsyncRequest = await AsyncRequest(
+        method=AsyncRequest.Method.POST,
         url=MOCK_REQUEST_URL,
         json={"name": "morpheus", "job": "leader"},
         validation_function=has_name,
@@ -492,6 +518,21 @@ class TestSanitizeForCSV:
         assert sanitize_list_for_csv([1, ["ab", "=de"]]) == [1, ["ab", "'=de"]]
 
 
+class TestValidateURL:
+    @pytest.mark.flaky
+    def test_valid_urls(self):
+        assert validate_url("https://www.gradio.app")
+        assert validate_url("http://gradio.dev")
+        assert validate_url(
+            "https://upload.wikimedia.org/wikipedia/commons/b/b0/Bengal_tiger_%28Panthera_tigris_tigris%29_female_3_crop.jpg"
+        )
+
+    def test_invalid_urls(self):
+        assert not (validate_url("C:/Users/"))
+        assert not (validate_url("C:\\Users\\"))
+        assert not (validate_url("/home/user"))
+
+
 class TestAppendUniqueSuffix:
     def test_no_suffix(self):
         name = "test"
@@ -509,5 +550,32 @@ class TestAppendUniqueSuffix:
         assert append_unique_suffix(name, list_of_names) == "test_4"
 
 
-if __name__ == "__main__":
-    unittest.main()
+@pytest.mark.parametrize(
+    "orig_filename, new_filename",
+    [
+        ("abc", "abc"),
+        ("$$AAabc&3", "AAabc3"),
+        ("$$AAabc&3", "AAabc3"),
+        ("$$AAa..b-c&3_", "AAa..b-c3_"),
+        ("$$AAa..b-c&3_", "AAa..b-c3_"),
+        (
+            "ゆかりです｡私､こんなかわいい服は初めて着ました…｡なんだかうれしくって､楽しいです｡歌いたくなる気分って､初めてです｡これがｱｲﾄﾞﾙってことなのかもしれませんね",
+            "ゆかりです私こんなかわいい服は初めて着ましたなんだかうれしくって楽しいです歌いたくなる気分って初めてですこれがｱｲﾄﾞﾙってことなの",
+        ),
+    ],
+)
+def test_strip_invalid_filename_characters(orig_filename, new_filename):
+    assert strip_invalid_filename_characters(orig_filename) == new_filename
+
+
+class TestAbspath:
+    def test_abspath_no_symlink(self):
+        resolved_path = str(abspath("../gradio/gradio/test_data/lion.jpg"))
+        assert ".." not in resolved_path
+
+    @mock.patch(
+        "pathlib.Path.is_symlink", return_value=True
+    )  # Have to patch since Windows doesn't allow creation of sym links without administrative privileges
+    def test_abspath_symlink(self, mock_islink):
+        resolved_path = str(abspath("../gradio/gradio/test_data/lion.jpg"))
+        assert ".." in resolved_path
